"""
Created on Thu Sep 28 15:17:50 2017

@author: zqwu
"""
import numpy as np
import tensorflow as tf
import copy
import sys

from deepchem.metrics import to_one_hot
from deepchem.models import KerasModel, layers
from deepchem.models.losses import L2Loss, SoftmaxCrossEntropy
from tensorflow.keras.layers import Input, Dense, Reshape, Softmax, Dropout, Conv1D, Concatenate, Lambda

# Common symbols in SMILES, note that Cl and Br are regarded as single symbol
default_dict = {
    '#': 1,
    '(': 2,
    ')': 3,
    '+': 4,
    '-': 5,
    '/': 6,
    '1': 7,
    '2': 8,
    '3': 9,
    '4': 10,
    '5': 11,
    '6': 12,
    '7': 13,
    '8': 14,
    '=': 15,
    'C': 16,
    'F': 17,
    'H': 18,
    'I': 19,
    'N': 20,
    'O': 21,
    'P': 22,
    'S': 23,
    '[': 24,
    '\\': 25,
    ']': 26,
    '_': 27,
    'c': 28,
    'Cl': 29,
    'Br': 30,
    'n': 31,
    'o': 32,
    's': 33
}


class TextCNNModel(KerasModel):
    """ A Convolutional neural network on smiles strings

    Reimplementation of the discriminator module in ORGAN [1]_ .
    Originated from [2]_.

    This model applies multiple 1D convolutional filters to
    the padded strings, then max-over-time pooling is applied on
    all filters, extracting one feature per filter.  All
    features are concatenated and transformed through several
    hidden layers to form predictions.

    This model is initially developed for sentence-level
    classification tasks, with words represented as vectors. In
    this implementation, SMILES strings are dissected into
    characters and transformed to one-hot vectors in a similar
    way. The model can be used for general molecular-level
    classification or regression tasks. It is also used in the
    ORGAN model as discriminator.

    Training of the model only requires SMILES strings input,
    all featurized datasets that include SMILES in the `ids`
    attribute are accepted. PDBbind, QM7 and QM7b are not
    supported. To use the model, `build_char_dict` should be
    called first before defining the model to build character
    dict of input dataset, example can be found in
    examples/delaney/delaney_textcnn.py

    References
    ----------
    .. [1]  Guimaraes, Gabriel Lima, et al. "Objective-reinforced generative adversarial networks (ORGAN) for sequence generation models." arXiv preprint arXiv:1705.10843 (2017).
    .. [2] Kim, Yoon. "Convolutional neural networks for sentence classification." arXiv preprint arXiv:1408.5882 (2014).

    """

    def __init__(self,
                 n_tasks,
                 char_dict,
                 seq_length,
                 n_embedding=75,
                 kernel_sizes=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 15, 20],
                 num_filters=[
                     100, 200, 200, 200, 200, 100, 100, 100, 100, 100, 160, 160
                 ],
                 dropout=0.25,
                 mode="classification",
                 **kwargs):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks
        char_dict: dict
            Mapping from characters in smiles to integers
        seq_length: int
            Length of sequences(after padding)
        n_embedding: int, optional
            Length of embedding vector
        filter_sizes: list of int, optional
            Properties of filters used in the conv net
        num_filters: list of int, optional
            Properties of filters used in the conv net
        dropout: float, optional
            Dropout rate
        mode: str
            Either "classification" or "regression" for type of model.
        """
        self.n_tasks = n_tasks
        self.char_dict = char_dict
        self.seq_length = max(seq_length, max(kernel_sizes))
        self.n_embedding = n_embedding
        self.kernel_sizes = kernel_sizes
        self.num_filters = num_filters
        self.dropout = dropout
        self.mode = mode

        # Build the model.

        smiles_seqs = Input(shape=(self.seq_length,), dtype=tf.int32)
        # Character embedding
        embedding = layers.DTNNEmbedding(
            n_embedding=self.n_embedding,
            periodic_table_length=len(self.char_dict.keys()) + 1)(smiles_seqs)
        pooled_outputs = []
        conv_layers = []
        for filter_size, num_filter in zip(self.kernel_sizes, self.num_filters):
            # Multiple convolutional layers with different filter widths
            conv_layers.append(
                Conv1D(kernel_size=filter_size,
                       filters=num_filter,
                       padding='valid')(embedding))
            # Max-over-time pooling
            reduced = Lambda(lambda x: tf.reduce_max(x, axis=1))(
                conv_layers[-1])
            pooled_outputs.append(reduced)
        # Concat features from all filters(one feature per filter)
        concat_outputs = Concatenate(axis=1)(pooled_outputs)
        dropout = Dropout(rate=self.dropout)(concat_outputs)
        dense = Dense(200, activation=tf.nn.relu)(dropout)
        # Highway layer from https://arxiv.org/pdf/1505.00387.pdf
        gather = layers.Highway()(dense)

        if self.mode == "classification":
            logits = Dense(self.n_tasks * 2)(gather)
            logits = Reshape((self.n_tasks, 2))(logits)
            output = Softmax()(logits)
            outputs = [output, logits]
            output_types = ['prediction', 'loss']
            loss = SoftmaxCrossEntropy()

        else:
            output = Dense(self.n_tasks * 1)(gather)
            output = Reshape((self.n_tasks, 1))(output)
            outputs = [output]
            output_types = ['prediction']
            loss = L2Loss()

        model = tf.keras.Model(inputs=[smiles_seqs], outputs=outputs)
        super(TextCNNModel, self).__init__(model,
                                           loss,
                                           output_types=output_types,
                                           **kwargs)

    @staticmethod
    def build_char_dict(dataset, default_dict=default_dict):
        """ Collect all unique characters(in smiles) from the dataset.
        This method should be called before defining the model to build appropriate char_dict
        """
        # SMILES strings
        X = dataset.ids
        # Maximum length is expanded to allow length variation during train and inference
        seq_length = int(max([len(smile) for smile in X]) * 1.2)
        # '_' served as delimiter and padding
        all_smiles = '_'.join(X)
        tot_len = len(all_smiles)
        # Initialize common characters as keys
        keys = list(default_dict.keys())
        out_dict = copy.deepcopy(default_dict)
        current_key_val = len(keys) + 1
        # Include space to avoid extra keys
        keys.extend([' '])
        extra_keys = []
        i = 0
        while i < tot_len:
            # For 'Cl', 'Br', etc.
            if all_smiles[i:i + 2] in keys:
                i = i + 2
            elif all_smiles[i:i + 1] in keys:
                i = i + 1
            else:
                # Character not recognized, add to extra_keys
                extra_keys.append(all_smiles[i])
                keys.append(all_smiles[i])
                i = i + 1
        # Add all extra_keys to char_dict
        for extra_key in extra_keys:
            out_dict[extra_key] = current_key_val
            current_key_val += 1
        return out_dict, seq_length

    @staticmethod
    def convert_bytes_to_char(s):
        s = ''.join(chr(b) for b in s)
        return s

    def smiles_to_seq_batch(self, ids_b):
        """Converts SMILES strings to np.array sequence.

        A tf.py_func wrapper is written around this when creating the input_fn for make_estimator
        """
        if isinstance(ids_b[0], bytes) and sys.version_info[
                0] != 2:  # Python 2.7 bytes and string are analogous
            ids_b = [
                TextCNNModel.convert_bytes_to_char(smiles) for smiles in ids_b
            ]
        smiles_seqs = [self.smiles_to_seq(smiles) for smiles in ids_b]
        smiles_seqs = np.vstack(smiles_seqs)
        return smiles_seqs

    def default_generator(self,
                          dataset,
                          epochs=1,
                          mode='fit',
                          deterministic=True,
                          pad_batches=True):
        """Transfer smiles strings to fixed length integer vectors"""
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if y_b is not None:
                    if self.mode == 'classification':
                        y_b = to_one_hot(y_b.flatten(),
                                         2).reshape(-1, self.n_tasks, 2)
                # Transform SMILES sequence to integers
                X_b = self.smiles_to_seq_batch(ids_b)
                yield ([X_b], [y_b], [w_b])

    def smiles_to_seq(self, smiles):
        """ Tokenize characters in smiles to integers
        """
        smiles_len = len(smiles)
        seq = [0]
        keys = self.char_dict.keys()
        i = 0
        while i < smiles_len:
            # Skip all spaces
            if smiles[i:i + 1] == ' ':
                i = i + 1
            # For 'Cl', 'Br', etc.
            elif smiles[i:i + 2] in keys:
                seq.append(self.char_dict[smiles[i:i + 2]])
                i = i + 2
            elif smiles[i:i + 1] in keys:
                seq.append(self.char_dict[smiles[i:i + 1]])
                i = i + 1
            else:
                raise ValueError('character not found in dict')
        for i in range(self.seq_length - len(seq)):
            # Padding with '_'
            seq.append(self.char_dict['_'])
        return np.array(seq, dtype=np.int32)


#################### Deprecation warnings for renamed TensorGraph models ####################  # noqa: E266

import warnings  # noqa: E402

TENSORGRAPH_DEPRECATION = "{} is deprecated and has been renamed to {} and will be removed in DeepChem 3.0."


class TextCNNTensorGraph(TextCNNModel):

    def __init__(self, *args, **kwargs):
        warnings.warn(
            TENSORGRAPH_DEPRECATION.format("TextCNNTensorGraph",
                                           "TextCNNModel"), FutureWarning)

        super(TextCNNTensorGraph, self).__init__(*args, **kwargs)



import numpy as np
import unittest
from deepchem.utils import voxel_utils
from deepchem.utils import hash_utils


class TestVoxelUtils(unittest.TestCase):

    def test_convert_atom_to_voxel(self):
        N = 5
        coordinates = np.random.rand(N, 3)
        atom_index = 2
        box_width = 16
        voxel_width = 1
        indices = voxel_utils.convert_atom_to_voxel(coordinates, atom_index,
                                                    box_width, voxel_width)
        assert indices.shape == (3,)

    def test_convert_pair_atom_to_voxel(self):
        N = 5
        M = 6
        coordinates1 = np.random.rand(N, 3)
        coordinates2 = np.random.rand(M, 3)
        atom_index_pair = (2, 3)
        box_width = 16
        voxel_width = 1
        indices = voxel_utils.convert_atom_pair_to_voxel(
            [coordinates1, coordinates2], atom_index_pair, box_width,
            voxel_width)
        assert indices.shape == (2, 3)

    def test_voxelize_convert_atom(self):
        N = 5
        coordinates = np.random.rand(N, 3)
        box_width = 16
        voxel_width = 1
        voxels_per_edge = int(box_width / voxel_width)
        get_voxels = voxel_utils.convert_atom_to_voxel
        hash_function = hash_utils.hash_ecfp
        feature_dict = {1: "C", 2: "CC"}
        nb_channel = 16
        features = voxel_utils.voxelize(get_voxels,
                                        coordinates,
                                        box_width,
                                        voxel_width,
                                        hash_function,
                                        feature_dict,
                                        nb_channel=nb_channel)
        assert features.shape == (voxels_per_edge, voxels_per_edge,
                                  voxels_per_edge, nb_channel)

    def test_voxelize_convert_atom_pair(self):
        N = 5
        M = 6
        coordinates1 = np.random.rand(N, 3)
        coordinates2 = np.random.rand(M, 3)
        coordinates = [coordinates1, coordinates2]
        box_width = 16
        voxel_width = 1
        voxels_per_edge = int(box_width / voxel_width)
        get_voxels = voxel_utils.convert_atom_pair_to_voxel
        hash_function = hash_utils.hash_ecfp_pair
        feature_dict = {(1, 2): ("C", "O"), (2, 3): ("CC", "OH")}
        nb_channel = 16
        features = voxel_utils.voxelize(get_voxels,
                                        coordinates,
                                        box_width,
                                        voxel_width,
                                        hash_function,
                                        feature_dict,
                                        nb_channel=nb_channel)
        assert features.shape == (voxels_per_edge, voxels_per_edge,
                                  voxels_per_edge, nb_channel)



import numpy as np

from deepchem.utils.typing import PymatgenComposition
from deepchem.feat import MaterialCompositionFeaturizer
from typing import Any


class ElementPropertyFingerprint(MaterialCompositionFeaturizer):
    """
    Fingerprint of elemental properties from composition.

    Based on the data source chosen, returns properties and statistics
    (min, max, range, mean, standard deviation, mode) for a compound
    based on elemental stoichiometry. E.g., the average electronegativity
    of atoms in a crystal structure. The chemical fingerprint is a
    vector of these statistics. For a full list of properties and statistics,
    see ``matminer.featurizers.composition.ElementProperty(data_source).feature_labels()``.

    This featurizer requires the optional dependencies pymatgen and
    matminer. It may be useful when only crystal compositions are available
    (and not 3D coordinates).

    See references [1]_, [2]_, [3]_, [4]_ for more details.

    References
    ----------
    .. [1] MagPie data: Ward, L. et al. npj Comput Mater 2, 16028 (2016).
        https://doi.org/10.1038/npjcompumats.2016.28
    .. [2] Deml data: Deml, A. et al. Physical Review B 93, 085142 (2016).
        10.1103/PhysRevB.93.085142
    .. [3] Matminer: Ward, L. et al. Comput. Mater. Sci. 152, 60-69 (2018).
    .. [4] Pymatgen: Ong, S.P. et al. Comput. Mater. Sci. 68, 314-319 (2013).

    Examples
    --------
    >>> import deepchem as dc
    >>> import pymatgen as mg
    >>> comp = mg.core.Composition("Fe2O3")
    >>> featurizer = dc.feat.ElementPropertyFingerprint()
    >>> features = featurizer.featurize([comp])
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape
    (65,)

    Note
    ----
    This class requires matminer and Pymatgen to be installed.
    `NaN` feature values are automatically converted to 0 by this featurizer.
    """

    def __init__(self, data_source: str = 'matminer'):
        """
        Parameters
        ----------
        data_source: str of "matminer", "magpie" or "deml" (default "matminer")
            Source for element property data.
        """
        self.data_source = data_source
        self.ep_featurizer: Any = None

    def _featurize(self, datapoint: PymatgenComposition,
                   **kwargs) -> np.ndarray:
        """
        Calculate chemical fingerprint from crystal composition.

        Parameters
        ----------
        datapoint: pymatgen.core.Composition object
            Composition object.

        Returns
        -------
        feats: np.ndarray
            Vector of properties and statistics derived from chemical
            stoichiometry. Some values may be NaN.
        """
        if 'composition' in kwargs and datapoint is None:
            datapoint = kwargs.get("composition")
            raise DeprecationWarning(
                'Composition is being phased out as a parameter, please pass "datapoint" instead.'
            )

        if self.ep_featurizer is None:
            try:
                from matminer.featurizers.composition import ElementProperty
                self.ep_featurizer = ElementProperty.from_preset(
                    self.data_source)
            except ModuleNotFoundError:
                raise ImportError(
                    "This class requires matminer to be installed.")

        try:
            feats = self.ep_featurizer.featurize(datapoint)
        except:
            feats = []

        return np.nan_to_num(np.array(feats))



import numpy as np
from typing import List, Any
from numpy.typing import ArrayLike
from deepchem.feat.graph_data import BatchGraphData

try:
    import torch
except ModuleNotFoundError:
    pass


def _get_atom_scopes(graph_index: ArrayLike) -> List[List[int]]:
    """Atom scope is a list of tuples with a single entry for every
    molecule in the batched graph. The entry indicates the beginning
    node index for a molecule and the number of nodes in the molecule.

    Parameters
    ----------
    graph_index: np.array
        An array containing a mapping between node index and the graph
    in the batched graph.

    Returns
    -------
    scopes: List[List[int]]
        Node index scope for each molecule in the batched graph.

    Example
    -------
    >>> import numpy as np
    >>> graph_index = np.array([0, 0, 1, 1, 1])
    >>> _get_atom_scopes(graph_index)
    [[0, 2], [2, 3]]
    """
    # graph_index indicates which atom belongs to which molecule
    mols = np.unique(graph_index)
    scopes = []
    for mol in mols:
        positions = np.where(graph_index == mol, 1, 0)
        scopes.append(
            [int(np.argmax(positions)),
             int(np.count_nonzero(positions))])
    return scopes


def _get_bond_scopes(edge_index: ArrayLike,
                     graph_index: ArrayLike) -> List[List[int]]:
    """Bond scope is a list of tuples with a single entry for every molecule
    in the batched graph. The entry indicates the beginning bond index for a
    molecule and the number of bonds in the molecule.

    Parameters
    ----------
    edge_index: np.array
        Graph connectivity in COO format with shape [2, num_edges]
    graph_index: np.array
        An array containing a mapping between node index and the graph
    in the batched graph.

    Returns
    -------
    scopes: List[List[int]]
        Bond index scope for each molecule in the batched graph.

    Example
    -------
    >>> edge_index = np.array([[0, 1, 2, 4], [1, 0, 4, 2]]) # a molecule with 4 bonds
    >>> graph_index = np.array([0, 0, 1, 1, 1])
    >>> _get_bond_scopes(edge_index, graph_index)
    [[0, 2], [2, 2]]
    """
    mols = np.unique(graph_index)
    bond_index = graph_index[edge_index[0]]  # type: ignore
    scopes = []
    for mol in mols:
        positions = np.where(bond_index == mol, 1, 0)
        scopes.append(
            [int(np.argmax(positions)),
             int(np.count_nonzero(positions))])
    return scopes


def _compute_b2revb(edge_index: np.ndarray) -> List[int]:
    """Every edge in a grover graph is a directed edge. Hence, a bond
    is represented by two edges of opposite directions. b2revb is a representation
    which stores for every edge, the index of reverse edge of that edge.

    Parameters
    ----------
    edge_index: np.array
        Graph connectivity in COO format with shape [2, num_edges]

    Returns
    -------
    b2revb: List[int]
        A mapping where an element at an index contains the index of the reverse bond.

    Example
    -------
    >>> import numpy as np
    >>> edge_index = np.array([[0, 1, 2, 4], [1, 0, 4, 2]])
    >>> _compute_b2revb(edge_index)
    [1, 0, 3, 2]
    """
    b2revb = [0] * edge_index.shape[1]
    for i, bond in enumerate(edge_index.T):
        for j, (sa, da) in enumerate(edge_index.T):
            if sa == bond[1] and da == bond[0]:
                b2revb[i] = j
    return b2revb


def _get_a2b(n_atoms: int, edge_index: np.ndarray) -> np.ndarray:
    """a2b is a mapping between atoms and their incoming bonds.

    Parameters
    ----------
    n_atoms: int
        Number of atoms
    edge_index: np.array
        Graph connectivity in COO format with shape [2, num_edges]

    Returns
    -------
    a2b: ArrayLike
        A mapping between atoms and their incoming bonds

    Example
    -------
    >>> import numpy as np
    >>> edge_index = np.array([[0, 1], [1, 2]])
    >>> n_atoms = 3
    >>> _get_a2b(n_atoms, edge_index)
    array([[0],
           [0],
           [1]])
    """
    a2b: List[List[Any]] = [[] for atom in range(n_atoms)]

    for i, bond in enumerate(edge_index.T):
        dest_atom = bond[1]
        a2b[dest_atom].append(i)

    # padding
    max_num_bonds = max(map(lambda x: len(x), a2b))
    atom_bond_mapping = np.asarray(
        [a2b[a] + [0] * (max_num_bonds - len(a2b[a])) for a in range(n_atoms)])

    return atom_bond_mapping


def extract_grover_attributes(molgraph: BatchGraphData):
    """Utility to extract grover attributes for grover model

    Parameters
    ----------
    molgraph: BatchGraphData
        A batched graph data representing a collection of molecules.

    Returns
    -------
    graph_attributes: Tuple
        A tuple containing atom features, bond features, atom to bond mapping, bond to atom mapping, bond to reverse bond mapping, atom to atom mapping, atom scope, bond scope, functional group labels and other additional features.

    Example
    -------
    >>> import deepchem as dc
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> smiles = ['CC', 'CCC', 'CC(=O)C']
    >>> featurizer = dc.feat.GroverFeaturizer(features_generator=dc.feat.CircularFingerprint())
    >>> graphs = featurizer.featurize(smiles)
    >>> molgraph = BatchGraphData(graphs)
    >>> attributes = extract_grover_attributes(molgraph)
    """
    fg_labels = getattr(molgraph, 'fg_labels')
    additional_features = getattr(molgraph, 'additional_features')
    f_atoms = molgraph.node_features
    f_bonds = molgraph.edge_features
    graph_index = molgraph.graph_index
    edge_index = molgraph.edge_index

    a_scope = _get_atom_scopes(graph_index)
    b_scope = _get_bond_scopes(edge_index, graph_index)
    b2revb = _compute_b2revb(edge_index)

    # computing a2b
    a2b = _get_a2b(molgraph.num_nodes, edge_index)

    f_atoms_tensor = torch.FloatTensor(f_atoms)
    f_bonds_tensor = torch.FloatTensor(f_bonds)
    fg_labels_tensor = torch.FloatTensor(fg_labels)
    additional_features_tensor = torch.FloatTensor(additional_features)
    a2b_tensor = torch.LongTensor(a2b)
    b2a_tensor = torch.LongTensor(molgraph.edge_index[0])
    b2revb_tensor = torch.LongTensor(b2revb)
    # only needed if using atom messages
    a2a = b2a_tensor[a2b_tensor]  # type: ignore
    a_scope_tensor = torch.LongTensor(np.asarray(a_scope))
    b_scope_tensor = torch.LongTensor(np.asarray(b_scope))
    return f_atoms_tensor, f_bonds_tensor, a2b_tensor, b2a_tensor, b2revb_tensor, a2a, a_scope_tensor, b_scope_tensor, fg_labels_tensor, additional_features_tensor



import numpy as np
import copy
import random
import deepchem


class TicTacToeEnvironment(deepchem.rl.Environment):
    """
    Play tictactoe against a randomly acting opponent
    """
    X = np.array([1.0, 0.0])
    O = np.array([0.0, 1.0])  # noqa: E741
    EMPTY = np.array([0.0, 0.0])

    ILLEGAL_MOVE_PENALTY = -3.0
    LOSS_PENALTY = -3.0
    NOT_LOSS = 0.1
    DRAW_REWARD = 5.0
    WIN_REWARD = 10.0

    def __init__(self):
        super(TicTacToeEnvironment, self).__init__([(3, 3, 2)], 9)
        self.reset()

    def reset(self):
        self._terminated = False
        self._state = [np.zeros(shape=(3, 3, 2), dtype=np.float32)]

        # Randomize who goes first
        if random.randint(0, 1) == 1:
            move = self.get_O_move()
            self._state[0][move[0]][move[1]] = TicTacToeEnvironment.O

    def step(self, action):
        self._state = copy.deepcopy(self._state)
        row = action // 3
        col = action % 3

        # Illegal move -- the square is not empty
        if not np.all(self._state[0][row][col] == TicTacToeEnvironment.EMPTY):
            self._terminated = True
            return TicTacToeEnvironment.ILLEGAL_MOVE_PENALTY

        # Move X
        self._state[0][row][col] = TicTacToeEnvironment.X

        # Did X Win
        if self.check_winner(TicTacToeEnvironment.X):
            self._terminated = True
            return TicTacToeEnvironment.WIN_REWARD

        if self.game_over():
            self._terminated = True
            return TicTacToeEnvironment.DRAW_REWARD

        move = self.get_O_move()
        self._state[0][move[0]][move[1]] = TicTacToeEnvironment.O

        # Did O Win
        if self.check_winner(TicTacToeEnvironment.O):
            self._terminated = True
            return TicTacToeEnvironment.LOSS_PENALTY

        if self.game_over():
            self._terminated = True
            return TicTacToeEnvironment.DRAW_REWARD
        return TicTacToeEnvironment.NOT_LOSS

    def get_O_move(self):
        empty_squares = []
        for row in range(3):
            for col in range(3):
                if np.all(
                        self._state[0][row][col] == TicTacToeEnvironment.EMPTY):
                    empty_squares.append((row, col))
        return random.choice(empty_squares)

    def check_winner(self, player):
        for i in range(3):
            row = np.sum(self._state[0][i][:], axis=0)
            if np.all(row == player * 3):
                return True
            col = np.sum(self._state[0][:][i], axis=0)
            if np.all(col == player * 3):
                return True

        diag1 = self._state[0][0][0] + self._state[0][1][1] + self._state[0][2][
            2]
        if np.all(diag1 == player * 3):
            return True
        diag2 = self._state[0][0][2] + self._state[0][1][1] + self._state[0][2][
            0]
        if np.all(diag2 == player * 3):
            return True
        return False

    def game_over(self):
        for i in range(3):
            for j in range(3):
                if np.all(self._state[0][i][j] == TicTacToeEnvironment.EMPTY):
                    return False
        return True

    def display(self):
        state = self._state[0]
        s = ""
        for row in range(3):
            for col in range(3):
                if np.all(state[row][col] == TicTacToeEnvironment.EMPTY):
                    s += "_"
                if np.all(state[row][col] == TicTacToeEnvironment.X):
                    s += "X"
                if np.all(state[row][col] == TicTacToeEnvironment.O):
                    s += "O"
            s += "\n"
        return s



"""
Utilities to score protein-ligand poses using DeepChem.
"""
import numpy as np


def pairwise_distances(coords1: np.ndarray, coords2: np.ndarray) -> np.ndarray:
    """Returns matrix of pairwise Euclidean distances.

    Parameters
    ----------
    coords1: np.ndarray
        A numpy array of shape `(N, 3)`
    coords2: np.ndarray
        A numpy array of shape `(M, 3)`

    Returns
    -------
    np.ndarray
        A `(N,M)` array with pairwise distances.
    """
    return np.sum((coords1[None, :] - coords2[:, None])**2, -1)**0.5


def cutoff_filter(d: np.ndarray, x: np.ndarray, cutoff=8.0) -> np.ndarray:
    """Applies a cutoff filter on pairwise distances

    Parameters
    ----------
    d: np.ndarray
        Pairwise distances matrix. A numpy array of shape `(N, M)`
    x: np.ndarray
        Matrix of shape `(N, M)`
    cutoff: float, optional (default 8)
        Cutoff for selection in Angstroms

    Returns
    -------
    np.ndarray
        A `(N,M)` array with values where distance is too large thresholded to 0.
    """
    return np.where(d < cutoff, x, np.zeros_like(x))


def vina_nonlinearity(c: np.ndarray, w: float, Nrot: int) -> np.ndarray:
    """Computes non-linearity used in Vina.

    Parameters
    ----------
    c: np.ndarray
        A numpy array of shape `(N, M)`
    w: float
        Weighting term
    Nrot: int
        Number of rotatable bonds in this molecule

    Returns
    -------
    np.ndarray
        A `(N, M)` array with activations under a nonlinearity.
    """
    out_tensor = c / (1 + w * Nrot)
    return out_tensor


def vina_repulsion(d: np.ndarray) -> np.ndarray:
    """Computes Autodock Vina's repulsion interaction term.

    Parameters
    ----------
    d: np.ndarray
        A numpy array of shape `(N, M)`.

    Returns
    -------
    np.ndarray
        A `(N, M)` array with repulsion terms.
    """
    return np.where(d < 0, d**2, np.zeros_like(d))


def vina_hydrophobic(d: np.ndarray) -> np.ndarray:
    """Computes Autodock Vina's hydrophobic interaction term.

    Here, d is the set of surface distances as defined in [1]_

    Parameters
    ----------
    d: np.ndarray
        A numpy array of shape `(N, M)`.

    Returns
    -------
    np.ndarray
        A `(N, M)` array of hydrophoboic interactions in a piecewise linear curve.

    References
    ----------
    .. [1] Jain, Ajay N. "Scoring noncovalent protein-ligand interactions:
        a continuous differentiable function tuned to compute binding affinities."
        Journal of computer-aided molecular design 10.5 (1996): 427-440.
    """
    out_tensor = np.where(d < 0.5, np.ones_like(d),
                          np.where(d < 1.5, 1.5 - d, np.zeros_like(d)))
    return out_tensor


def vina_hbond(d: np.ndarray) -> np.ndarray:
    """Computes Autodock Vina's hydrogen bond interaction term.

    Here, d is the set of surface distances as defined in [1]_

    Parameters
    ----------
    d: np.ndarray
        A numpy array of shape `(N, M)`.

    Returns
    -------
    np.ndarray
        A `(N, M)` array of hydrophoboic interactions in a piecewise linear curve.

    References
    ----------
    .. [1] Jain, Ajay N. "Scoring noncovalent protein-ligand interactions:
        a continuous differentiable function tuned to compute binding affinities."
        Journal of computer-aided molecular design 10.5 (1996): 427-440.
    """
    out_tensor = np.where(
        d < -0.7, np.ones_like(d),
        np.where(d < 0, (1.0 / 0.7) * (0 - d), np.zeros_like(d)))
    return out_tensor


def vina_gaussian_first(d: np.ndarray) -> np.ndarray:
    """Computes Autodock Vina's first Gaussian interaction term.

    Here, d is the set of surface distances as defined in [1]_

    Parameters
    ----------
    d: np.ndarray
        A numpy array of shape `(N, M)`.

    Returns
    -------
    np.ndarray
        A `(N, M)` array of gaussian interaction terms.

    References
    ----------
    .. [1] Jain, Ajay N. "Scoring noncovalent protein-ligand interactions:
        a continuous differentiable function tuned to compute binding affinities."
        Journal of computer-aided molecular design 10.5 (1996): 427-440.
    """
    out_tensor = np.exp(-(d / 0.5)**2)
    return out_tensor


def vina_gaussian_second(d: np.ndarray) -> np.ndarray:
    """Computes Autodock Vina's second Gaussian interaction term.

    Here, d is the set of surface distances as defined in [1]_

    Parameters
    ----------
    d: np.ndarray
        A numpy array of shape `(N, M)`.

    Returns
    -------
    np.ndarray
        A `(N, M)` array of gaussian interaction terms.

    References
    ----------
    .. [1] Jain, Ajay N. "Scoring noncovalent protein-ligand interactions:
        a continuous differentiable function tuned to compute binding affinities."
        Journal of computer-aided molecular design 10.5 (1996): 427-440.
    """
    out_tensor = np.exp(-((d - 3) / 2)**2)
    return out_tensor


def weighted_linear_sum(w: np.ndarray, x: np.ndarray) -> np.ndarray:
    """Computes weighted linear sum.

    Parameters
    ----------
    w: np.ndarray
        A numpy array of shape `(N,)`
    x: np.ndarray
        A numpy array of shape `(N, M, L)`

    Returns
    -------
    np.ndarray
        A numpy array of shape `(M, L)`
    """
    return np.tensordot(w, x, axes=1)


def vina_energy_term(coords1: np.ndarray, coords2: np.ndarray,
                     weights: np.ndarray, wrot: float, Nrot: int) -> np.ndarray:
    """Computes the Vina Energy function for two molecular conformations

    Parameters
    ----------
    coords1: np.ndarray
        Molecular coordinates of shape `(N, 3)`
    coords2: np.ndarray
        Molecular coordinates of shape `(M, 3)`
    weights: np.ndarray
        A numpy array of shape `(5,)`. The 5 values are weights for repulsion interaction term,
        hydrophobic interaction term, hydrogen bond interaction term,
        first Gaussian interaction term and second Gaussian interaction term.
    wrot: float
        The scaling factor for nonlinearity
    Nrot: int
        Number of rotatable bonds in this calculation

    Returns
    -------
    np.ndarray
        A scalar value with free energy
    """
    # TODO(rbharath): The autodock vina source computes surface distances
    # which take into account the van der Waals radius of each atom type.
    dists = pairwise_distances(coords1, coords2)
    repulsion = vina_repulsion(dists)
    hydrophobic = vina_hydrophobic(dists)
    hbond = vina_hbond(dists)
    gauss_1 = vina_gaussian_first(dists)
    gauss_2 = vina_gaussian_second(dists)

    # Shape (N, M)
    interactions = weighted_linear_sum(
        weights, np.array([repulsion, hydrophobic, hbond, gauss_1, gauss_2]))

    # Shape (N, M)
    thresholded = cutoff_filter(dists, interactions)

    free_energies = vina_nonlinearity(thresholded, wrot, Nrot)
    return np.sum(free_energies)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import GRU, Linear, ReLU, Sequential
from typing import Iterable, List, Tuple, Optional, Dict, Literal
from deepchem.metrics import to_one_hot

import deepchem as dc
from deepchem.models.losses import SoftmaxCrossEntropy
from deepchem.feat.graph_data import BatchGraphData
from deepchem.models.losses import (
    GlobalMutualInformationLoss,
    LocalMutualInformationLoss,
)
from deepchem.models.torch_models.layers import MultilayerPerceptron
from deepchem.models.torch_models.modular import ModularTorchModel

try:
    from torch_geometric.nn import GINConv, NNConv, global_add_pool
    from torch_geometric.nn.aggr import Set2Set
except ImportError:
    pass


class GINEncoder(torch.nn.Module):
    """
    Graph Information Network (GIN) encoder. This is a graph convolutional network that produces encoded representations for molecular graph inputs.

    Parameters
    ----------
    num_features: int
        The number of node features
    embedding_dim: int
        The dimension of the output embedding
    num_gc_layers: int, optional (default 5)
        The number of graph convolutional layers to use

    Example
    -------
    >>> import numpy as np
    >>> from deepchem.models.torch_models.infograph import GINEncoder
    >>> from deepchem.feat.graph_data import GraphData
    >>> encoder = GINEncoder(num_features=25, embedding_dim=32)
    >>> node_features = np.random.randn(10, 25)
    >>> edge_index = np.array([[0, 1, 2], [1, 2, 3]])
    >>> edge_features = np.random.randn(3, 10)
    >>> graph_index = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    >>> data = GraphData(node_features=node_features, edge_index=edge_index, edge_features=edge_features, graph_index=graph_index).numpy_to_torch()
    >>> embedding, intermediate_embeddings = encoder(data)
    >>> print(embedding.shape)
    torch.Size([1, 30])

    References
    ----------
    .. [1] Xu, K., Hu, W., Leskovec, J. & Jegelka, S. How Powerful are Graph Neural Networks? arXiv:1810.00826 [cs, stat] (2019).

    """

    def __init__(self,
                 num_features: int,
                 embedding_dim: int,
                 num_gc_layers: int = 5):
        dim = int(
            embedding_dim / num_gc_layers
        )  # the output dimension of this encoder is modified by the number of GC layers, so this is necessary to ensure that the output dimension is consistent with the InfoGraphEncoder
        super().__init__()

        self.num_gc_layers = num_gc_layers

        self.convs = torch.nn.ModuleList()
        self.bns = torch.nn.ModuleList()

        for i in range(num_gc_layers):
            if i == 0:
                nn = Sequential(Linear(num_features, dim), ReLU(),
                                Linear(dim, dim))
            else:
                nn = Sequential(Linear(dim, dim), ReLU(), Linear(dim, dim))
            conv = GINConv(nn)
            bn = torch.nn.BatchNorm1d(dim)

            self.convs.append(conv)
            self.bns.append(bn)

    def forward(self, data):
        """
        Encodes the input graph data.

        Parameters
        ----------
        data : BatchGraphData
            The batched input graph data.

        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor]
            A tuple containing the encoded representation and intermediate embeddings.
        """
        xs = []
        x = data.node_features
        for i in range(self.num_gc_layers):

            x = F.relu(self.convs[i](x, data.edge_index))
            x = self.bns[i](x)
            xs.append(x)

        xpool = [global_add_pool(x, data.graph_index) for x in xs]
        x = torch.cat(xpool, 1)
        xs = torch.cat(xs, 1)
        return x, xs


class InfoGraphEncoder(torch.nn.Module):
    """
    The encoder for the InfoGraph model. It is a message passing graph convolutional
    network that produces encoded representations for molecular graph inputs.

    Parameters
    ----------
    num_features: int
        Number of node features for each input
    edge_features: int
        Number of edge features for each input
    embedding_dim: int
        Dimension of the embedding

    Example
    -------
    >>> import numpy as np
    >>> from deepchem.models.torch_models.infograph import InfoGraphEncoder
    >>> from deepchem.feat.graph_data import GraphData
    >>> encoder = InfoGraphEncoder(num_features=25, edge_features=10, embedding_dim=32)
    >>> node_features = np.random.randn(10, 25)
    >>> edge_index = np.array([[0, 1, 2], [1, 2, 3]])
    >>> edge_features = np.random.randn(3, 10)
    >>> graph_index = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    >>> data = GraphData(node_features=node_features, edge_index=edge_index, edge_features=edge_features, graph_index=graph_index).numpy_to_torch()
    >>> embedding, feature_map = encoder(data)
    >>> print(embedding.shape)
    torch.Size([1, 64])
    """

    def __init__(self, num_features, edge_features, embedding_dim):
        super().__init__()
        self.lin0 = torch.nn.Linear(num_features, embedding_dim)

        nn = Sequential(Linear(edge_features, 128), ReLU(),
                        Linear(128, embedding_dim * embedding_dim))
        self.conv = NNConv(embedding_dim,
                           embedding_dim,
                           nn,
                           aggr='mean',
                           root_weight=False)
        self.gru = GRU(embedding_dim, embedding_dim)

        self.set2set = Set2Set(embedding_dim, processing_steps=3)

    def forward(self, data):
        """
        Encode input graphs into an embedding and feature map.

        Parameters
        ----------
        data: Union[BatchGraphData, GraphData]
            Contains information about graphs.

        Returns
        -------
        torch.Tensor
            Encoded tensor of input data.
        torch.Tensor
            Feature map tensor of input data.

        """
        out = F.relu(self.lin0(data.node_features))
        h = out.unsqueeze(0)

        for i in range(3):
            m = F.relu(self.conv(out, data.edge_index, data.edge_features))
            out, h = self.gru(m.unsqueeze(0), h)
            out = out.squeeze(0)
            feat_map = out

        # set2set doubles the dimensionality of the embedding
        out = self.set2set(out, data.graph_index)
        return out, feat_map


class InfoGraph(nn.Module):
    """
    The nn.Module for InfoGraph. This class defines the forward pass of InfoGraph.

    References
    ----------
    1. Sun, F.-Y., Hoffmann, J., Verma, V. & Tang, J. InfoGraph: Unsupervised and Semi-supervised Graph-Level Representation Learning via Mutual Information Maximization. Preprint at http://arxiv.org/abs/1908.01000 (2020).

    Example
    -------
    >>> from deepchem.models.torch_models.infograph import InfoGraphModel
    >>> from deepchem.feat.molecule_featurizers import MolGraphConvFeaturizer
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> num_feat = 30
    >>> num_edge = 11
    >>> infographmodular = InfoGraphModel(num_feat, num_edge, 64)
    >>> smiles = ['C1=CC=CC=C1', 'C1=CC=CC=C1C2=CC=CC=C2']
    >>> featurizer = MolGraphConvFeaturizer(use_edges=True)
    >>> graphs = BatchGraphData(featurizer.featurize(smiles))
    >>> graphs = graphs.numpy_to_torch(infographmodular.device)
    >>> model = infographmodular.model
    >>> global_enc, local_enc = model(graphs)

    """

    def __init__(self, encoder, local_d, global_d, prior_d, init_emb=False):
        super().__init__()
        self.encoder = encoder
        self.local_d = local_d
        self.global_d = global_d
        self.prior_d = prior_d
        if init_emb:
            self.init_emb()

    def init_emb(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                torch.nn.init.xavier_uniform_(m.weight.data)
                if m.bias is not None:
                    m.bias.data.fill_(0.0)

    def forward(self, data):
        y, M = self.encoder(data)
        g_enc = self.global_d(y)
        l_enc = self.local_d(M)
        return g_enc, l_enc


class InfoGraphFinetune(nn.Module):
    """The finetuning module for InfoGraph model

    Parameters
    ----------
    encoder: nn.Module
        An encoder to encode input graph data
    fc1: nn.Module
        A fully connected layer
    fc2: nn.Module
        A fully connected layer

    Example
    -------
    >>> from deepchem.models.torch_models.infograph import InfoGraphModel
    >>> from deepchem.feat.molecule_featurizers import MolGraphConvFeaturizer
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> num_feat = 30
    >>> num_edge = 11
    >>> infographmodular = InfoGraphModel(num_feat, num_edge, num_gc_layers=1, task='regression', n_tasks=1)
    >>> smiles = ['C1=CC=CC=C1', 'C1=CC=CC=C1C2=CC=CC=C2']
    >>> featurizer = MolGraphConvFeaturizer(use_edges=True)
    >>> graphs = BatchGraphData(featurizer.featurize(smiles))
    >>> graphs = graphs.numpy_to_torch(infographmodular.device)
    >>> model = infographmodular.model
    >>> predictions = model(graphs)

    Reference
    ---------
    .. Sun, F.-Y, et. al, "InfoGraph: Unsupervised and Semi-supervised Graph-Level Representation Learning via Mutual Information Maximization".
    """

    def __init__(self, encoder, fc1, fc2, init_emb=False):
        super().__init__()
        self.encoder = encoder
        self.fc1 = fc1
        self.fc2 = fc2
        if init_emb:
            self.init_emb()

    def init_emb(self):
        for m in self.modules():
            if isinstance(m, nn.Linear):
                torch.nn.init.xavier_uniform_(m.weight.data)
                if m.bias is not None:
                    m.bias.data.fill_(0.0)

    def forward(self, data):
        y, _ = self.encoder(data)
        return self.fc2(F.relu(self.fc1(y)))


class InfoGraphModel(ModularTorchModel):
    """InfoGraphMode

    InfoGraphModel is a model which learn graph-level representation via unsupervised learning. To this end,
    the model aims to maximize the mutual information between the representations of entire graphs and the representations of substructures of different granularity (eg. nodes, edges, triangles)

    The unsupervised training of InfoGraph involves two encoders: one that encodes the entire graph and another that encodes substructures of different sizes. The mutual information between the two encoder outputs is maximized using a contrastive loss function.
    The model randomly samples pairs of graphs and substructures, and then maximizes their mutual information by minimizing their distance in a learned embedding space.

    This can be used for downstream tasks such as graph classification and molecular property prediction.It is implemented as a ModularTorchModel in order to facilitate transfer learning.

    References
    ----------
    1. Sun, F.-Y., Hoffmann, J., Verma, V. & Tang, J. InfoGraph: Unsupervised and Semi-supervised Graph-Level Representation Learning via Mutual Information Maximization. Preprint at http://arxiv.org/abs/1908.01000 (2020).

    Parameters
    ----------
    num_features: int
        Number of node features for each input
    edge_features: int
        Number of edge features for each input
    embedding_dim: int
        Dimension of the embedding
    num_gc_layers: int
        Number of graph convolutional layers
    prior: bool
        Whether to use a prior expectation in the loss term
    gamma: float
        Weight of the prior expectation in the loss term
    measure: str
        The divergence measure to use for the unsupervised loss. Options are 'GAN', 'JSD',
        'KL', 'RKL', 'X2', 'DV', 'H2', or 'W1'.
    average_loss: bool
        Whether to average the loss over the batch

    Example
    -------
    >>> from deepchem.models.torch_models.infograph import InfoGraphModel
    >>> from deepchem.feat import MolGraphConvFeaturizer
    >>> from deepchem.data import NumpyDataset
    >>> import torch
    >>> import tempfile
    >>> tempdir = tempfile.TemporaryDirectory()
    >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
    >>> featurizer = MolGraphConvFeaturizer(use_edges=True)
    >>> X = featurizer.featurize(smiles)
    >>> y = torch.randint(0, 2, size=(2, 1)).float()
    >>> w = torch.ones(size=(2, 1)).float()
    >>> dataset = NumpyDataset(X, y, w)
    >>> num_feat, edge_dim = 30, 11  # num feat and edge dim by molgraph conv featurizer
    >>> pretrain_model = InfoGraphModel(num_feat, edge_dim, num_gc_layers=1, task='pretraining', model_dir=tempdir.name)
    >>> pretraining_loss = pretrain_model.fit(dataset, nb_epoch=1)
    >>> pretrain_model.save_checkpoint()
    >>> finetune_model = InfoGraphModel(num_feat, edge_dim, num_gc_layers=1, task='regression', n_tasks=1, model_dir=tempdir.name)
    >>> finetune_model.restore(components=['encoder'])
    >>> finetuning_loss = finetune_model.fit(dataset)
    """

    def __init__(self,
                 num_features,
                 embedding_dim,
                 num_gc_layers=5,
                 prior=True,
                 gamma=.1,
                 measure='JSD',
                 average_loss=True,
                 task='pretraining',
                 n_tasks: Optional[int] = None,
                 **kwargs):
        if task == 'regression':
            assert n_tasks, 'Number of prediction tasks required for building regression model'
        self.num_features = num_features
        self.embedding_dim = embedding_dim * num_gc_layers
        self.num_gc_layers = num_gc_layers
        self.gamma = gamma
        self.prior = prior
        self.measure = measure
        self.average_loss = average_loss
        self.localloss = LocalMutualInformationLoss()._create_pytorch_loss(
            measure, average_loss)
        self.task = task
        self.n_tasks = n_tasks
        self.components = self.build_components()
        self.model = self.build_model()
        super().__init__(self.model, self.components, **kwargs)

    def build_components(self) -> dict:
        """
        Build the components of the model. InfoGraph is an unsupervised molecular graph representation learning model. It consists of an encoder, a local discriminator, a global discriminator, and a prior discriminator.

        The unsupervised loss is calculated by the mutual information in embedding representations at all layers.

        Components list, type and description:
        --------------------------------------
        encoder: GINEncoder, graph convolutional encoder

        local_d: MultilayerPerceptron, local discriminator

        global_d: MultilayerPerceptron, global discriminator

        prior_d: MultilayerPerceptron, prior discriminator
        fc1: MultilayerPerceptron, dense layer used during finetuning
        fc2: MultilayerPerceptron, dense layer used during finetuning
        """
        components: Dict[str, nn.Module] = {}
        if self.task == 'pretraining':
            components['encoder'] = GINEncoder(self.num_features,
                                               self.embedding_dim,
                                               self.num_gc_layers)
            components['local_d'] = MultilayerPerceptron(self.embedding_dim,
                                                         self.embedding_dim,
                                                         (self.embedding_dim,),
                                                         skip_connection=True)
            components['global_d'] = MultilayerPerceptron(self.embedding_dim,
                                                          self.embedding_dim,
                                                          (self.embedding_dim,),
                                                          skip_connection=True)
            components['prior_d'] = MultilayerPerceptron(
                self.embedding_dim,
                1, (self.embedding_dim,),
                activation_fn='sigmoid')
        elif self.task == 'regression':
            components['encoder'] = GINEncoder(self.num_features,
                                               self.embedding_dim,
                                               self.num_gc_layers)
            components['fc1'] = torch.nn.Linear(self.embedding_dim,
                                                self.embedding_dim)
            # n_tasks is Optional[int] while argument 2 of nn.Linear has to be of type int
            components['fc2'] = torch.nn.Linear(self.embedding_dim,
                                                self.n_tasks)  # type: ignore
        return components

    def build_model(self) -> nn.Module:
        if self.task == 'pretraining':
            model = InfoGraph(**self.components)
        elif self.task == 'regression':
            model = InfoGraphFinetune(**self.components)  # type: ignore
        return model

    def loss_func(self, inputs, labels, weights):
        if self.task == 'pretraining':
            y, M = self.components['encoder'](inputs)
            g_enc = self.components['global_d'](y)
            l_enc = self.components['local_d'](M)
            local_global_loss = self.localloss(l_enc, g_enc, inputs.graph_index)
            if self.prior:
                prior = torch.rand_like(y)
                term_a = torch.log(self.components['prior_d'](prior)).mean()
                term_b = torch.log(1.0 - self.components['prior_d'](y)).mean()
                prior = -(term_a + term_b) * self.gamma
            else:
                prior = 0
            return local_global_loss + prior
        elif self.task == 'regression':
            loss_fn = nn.MSELoss()
            y = self.model(inputs)
            return loss_fn(y, labels)

    def _prepare_batch(self, batch):
        """
        Prepares the batch for the model by converting the GraphData numpy arrays to torch tensors and moving them to the device.
        """
        inputs, labels, weights = batch
        inputs = BatchGraphData(inputs[0]).numpy_to_torch(self.device)

        _, labels, weights = super()._prepare_batch(([], labels, weights))

        if (len(labels) != 0) and (len(weights) != 0):
            labels = labels[0]
            weights = weights[0]

        return inputs, labels, weights

    def restore(  # type: ignore
            self,
            components: Optional[List[str]] = None,
            checkpoint: Optional[str] = None,
            model_dir: Optional[str] = None,
            map_location: Optional[torch.device] = None) -> None:
        if checkpoint is None:
            checkpoints = sorted(self.get_checkpoints(model_dir))
            if len(checkpoints) == 0:
                raise ValueError('No checkpoint found')
            checkpoint = checkpoints[0]
        data = torch.load(checkpoint, map_location=map_location)
        for name, state_dict in data.items():
            if name != 'model' and name in self.components.keys():
                self.components[name].load_state_dict(state_dict)

        self.build_model()


class InfoGraphStar(torch.nn.Module):
    """
    The nn.Module for InfoGraphStar. This class defines the forward pass of InfoGraphStar.

    Parameters
    ----------
    encoder: torch.nn.Module
        The encoder for InfoGraphStar.
    unsup_encoder: torch.nn.Module
        The unsupervised encoder for InfoGraph, of identical architecture to encoder.
    ff1: torch.nn.Module
        The first feedforward layer for InfoGraphStar.
    ff2: torch.nn.Module
        The second feedforward layer for InfoGraphStar.
    fc1: torch.nn.Module
        The first fully connected layer for InfoGraphStar.
    fc2: torch.nn.Module
        The second fully connected layer for InfoGraphStar.

    References
    ----------
    .. [1] F.-Y. Sun, J. Hoffmann, V. Verma, and J. Tang, “InfoGraph: Unsupervised and Semi-supervised Graph-Level Representation Learning via Mutual Maximization.” arXiv, Jan. 17, 2020. http://arxiv.org/abs/1908.01000


    Example
    -------
    >>> import torch
    >>> import numpy as np
    >>> from deepchem.models.torch_models.infograph import InfoGraphStarModel
    >>> from deepchem.feat.molecule_featurizers import MolGraphConvFeaturizer
    >>> from deepchem.feat.graph_data import BatchGraphData
    >>> num_feat = 30
    >>> num_edge = 11
    >>> infographmodular = InfoGraphStarModel(num_feat,num_edge,64)
    >>> smiles = ['C1=CC=CC=C1', 'C1=CC=CC=C1C2=CC=CC=C2']
    >>> featurizer = MolGraphConvFeaturizer(use_edges=True)
    >>> graphs = BatchGraphData(featurizer.featurize(smiles)).numpy_to_torch(infographmodular.device)
    >>> model = infographmodular.model
    >>> output = model(graphs).cpu().detach().numpy()

    """

    def __init__(self,
                 encoder,
                 unsup_encoder,
                 ff1,
                 ff2,
                 fc1,
                 fc2,
                 local_d,
                 global_d,
                 mode='regression',
                 num_tasks=1,
                 num_classes=2,
                 init_emb=False):
        super().__init__()
        self.encoder = encoder
        self.unsup_encoder = unsup_encoder
        self.ff1 = ff1
        self.ff2 = ff2
        self.fc1 = fc1
        self.fc2 = fc2
        self.local_d = local_d
        self.global_d = global_d
        self.mode = mode
        self.num_tasks = num_tasks
        self.num_classes = num_classes
        if init_emb:
            self.init_emb()

    def init_emb(self):
        """
        Initializes the parameters of the model by setting their values from a uniform distribution and filling the bias with 0.
        """
        for m in self.modules():
            if isinstance(m, nn.Linear):
                torch.nn.init.xavier_uniform_(m.weight.data)
                if m.bias is not None:
                    m.bias.data.fill_(0.0)

    def forward(self, data):
        """
        Forward pass for InfoGraphStar.

        Parameters
        ----------
        data: Union[GraphData, BatchGraphData]
            The input data, either a single graph or a batch of graphs.
        """

        out, M = self.encoder(data)
        out = F.relu(self.fc1(out))
        pred = self.fc2(out)
        if self.mode == 'classification':
            pred = torch.reshape(pred, (-1, self.num_tasks, self.num_classes))
        return pred


class InfoGraphStarModel(ModularTorchModel):
    """
    InfographStar is a semi-supervised graph convolutional network for predicting molecular properties.
    It aims to maximize the mutual information between the graph-level representation and the
    representations of substructures of different scales. It does this by producing graph-level
    encodings and substructure encodings, and then using a discriminator to classify if they
    are from the same molecule or not.

    Supervised training is done by using the graph-level encodings to predict the target property. Semi-supervised training is done by adding a loss term that maximizes the mutual information between the graph-level encodings and the substructure encodings to the supervised loss.
    These modes can be chosen by setting the training_mode parameter.

    To conduct training in unsupervised mode, use InfoGraphModel.

    References
    ----------
    .. [1] F.-Y. Sun, J. Hoffmann, V. Verma, and J. Tang, “InfoGraph: Unsupervised and Semi-supervised Graph-Level Representation Learning via Mutual Maximization.” arXiv, Jan. 17, 2020. http://arxiv.org/abs/1908.01000

    Parameters
    ----------
    num_features: int
        Number of node features for each input
    edge_features: int
        Number of edge features for each input
    embedding_dim: int
        Dimension of the embedding
    training_mode: str
        The mode to use for training. Options are 'supervised', 'semisupervised'. For unsupervised training, use InfoGraphModel.
    measure: str
        The divergence measure to use for the unsupervised loss. Options are 'GAN', 'JSD',
        'KL', 'RKL', 'X2', 'DV', 'H2', or 'W1'.
    average_loss: bool
        Whether to average the loss over the batch

    Examples
    --------
    >>> from deepchem.models.torch_models import InfoGraphStarModel
    >>> from deepchem.feat import MolGraphConvFeaturizer
    >>> from deepchem.data import NumpyDataset
    >>> import torch
    >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
    >>> featurizer = MolGraphConvFeaturizer(use_edges=True)
    >>> X = featurizer.featurize(smiles)
    >>> y = torch.randint(0, 2, size=(2, 1)).float()
    >>> w = torch.ones(size=(2, 1)).float()
    >>> ds = NumpyDataset(X, y, w)
    >>> num_feat = max([ds.X[i].num_node_features for i in range(len(ds))])
    >>> edge_dim = max([ds.X[i].num_edge_features for i in range(len(ds))])
    >>> model = InfoGraphStarModel(num_feat, edge_dim, 15, training_mode='semisupervised')
    >>> loss = model.fit(ds, nb_epoch=1)
    """

    def __init__(self,
                 num_features,
                 edge_features,
                 embedding_dim,
                 task: Literal['supervised', 'semisupervised'] = 'supervised',
                 mode: Literal['regression', 'classification'] = 'regression',
                 num_classes=2,
                 num_tasks=1,
                 measure='JSD',
                 average_loss=True,
                 num_gc_layers=5,
                 **kwargs):

        assert task in ['supervised', 'semisupervised'], 'Invalid model task'
        assert mode in ['regression', 'classification'], 'Invalid model mode'
        self.edge_features = edge_features
        self.local = True
        self.prior = False
        self.gamma = .1
        self.num_features = num_features
        self.task = task
        self.mode = mode
        self.num_classes = num_classes
        if self.mode == 'regression':
            self.output_dim = num_tasks
        elif self.mode == 'classification':
            self.num_tasks = num_tasks
            self.output_dim = num_classes * num_tasks
            self.class_loss = SoftmaxCrossEntropy()._create_pytorch_loss()
        if self.task == 'supervised':
            self.embedding_dim = embedding_dim
        elif self.task == 'semisupervised':
            self.embedding_dim = embedding_dim * num_gc_layers
            self.num_gc_layers = num_gc_layers
        self.localloss = LocalMutualInformationLoss()._create_pytorch_loss(
            measure, average_loss)
        self.globalloss = GlobalMutualInformationLoss()._create_pytorch_loss(
            measure, average_loss)

        self.components = self.build_components()
        self.model = self.build_model()
        super().__init__(self.model, self.components, **kwargs)

    def build_components(self):
        """
        Builds the components of the InfoGraphStar model. InfoGraphStar works by maximizing the mutual information between the graph-level representation and the representations of substructures of different scales.

        It does this by producing graph-level encodings and substructure encodings, and then using a discriminator to classify if they are from the same molecule or not.

        The encoder is a graph convolutional network that produces the graph-level encodings and substructure encodings.

        In a supervised training mode, only 1 encoder is used and the encodings are not compared, while in a semi-supvervised training mode they are different in order to prevent negative transfer from the pretraining stage.

        The local discriminator is a multilayer perceptron that classifies if the substructure encodings are from the same molecule or not while the global discriminator classifies if the graph-level encodings are from the same molecule or not.

        Components list, type and description:
        --------------------------------------
        encoder: InfoGraphEncoder

        unsup_encoder: InfoGraphEncoder for supervised or GINEncoder for unsupervised training

        ff1: MultilayerPerceptron, feedforward network

        ff2: MultilayerPerceptron, feedforward network

        fc1: torch.nn.Linear, fully connected layer

        fc2: torch.nn.Linear, fully connected layer

        local_d: MultilayerPerceptron, local discriminator

        global_d: MultilayerPerceptron, global discriminator
        """
        if self.task == 'supervised':
            return {
                'encoder':
                    InfoGraphEncoder(self.num_features, self.edge_features,
                                     self.embedding_dim),
                'unsup_encoder':
                    InfoGraphEncoder(self.num_features, self.edge_features,
                                     self.embedding_dim),
                'ff1':
                    MultilayerPerceptron(2 * self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,)),
                'ff2':
                    MultilayerPerceptron(2 * self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,)),
                'fc1':
                    torch.nn.Linear(2 * self.embedding_dim, self.embedding_dim),
                'fc2':
                    torch.nn.Linear(self.embedding_dim, self.output_dim),
                'local_d':
                    MultilayerPerceptron(self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,),
                                         skip_connection=True),
                'global_d':
                    MultilayerPerceptron(2 * self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,),
                                         skip_connection=True)
            }
        elif self.task == 'semisupervised':
            return {
                'encoder':
                    InfoGraphEncoder(self.num_features, self.edge_features,
                                     self.embedding_dim),
                'unsup_encoder':
                    GINEncoder(self.num_features, self.embedding_dim,
                               self.num_gc_layers),
                'ff1':
                    MultilayerPerceptron(2 * self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,)),
                'ff2':
                    MultilayerPerceptron(self.embedding_dim, self.embedding_dim,
                                         (self.embedding_dim,)),
                'fc1':
                    torch.nn.Linear(2 * self.embedding_dim, self.embedding_dim),
                'fc2':
                    torch.nn.Linear(self.embedding_dim, self.output_dim),
                'local_d':
                    MultilayerPerceptron(self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,),
                                         skip_connection=True),
                'global_d':
                    MultilayerPerceptron(self.embedding_dim,
                                         self.embedding_dim,
                                         (self.embedding_dim,),
                                         skip_connection=True)
            }

    def build_model(self):
        """
        Builds the InfoGraph model by unpacking the components dictionary and passing them to the InfoGraph nn.module.
        """
        if self.mode == 'regression':
            return InfoGraphStar(**self.components,)
        elif self.mode == 'classification':
            return InfoGraphStar(**self.components,
                                 mode=self.mode,
                                 num_tasks=self.num_tasks,
                                 num_classes=self.num_classes)

    def loss_func(self, inputs, labels, weights):
        sup_loss = self.sup_loss(inputs, labels)
        if self.task == 'semisupervised':
            local_unsup_loss = self.local_unsup_loss(inputs)
            global_unsup_loss = self.global_unsup_loss(inputs, labels, weights)
            loss = sup_loss + local_unsup_loss + global_unsup_loss * self.learning_rate
            # original implementation also includes an option if not using a separate encoder:
            # loss = sup_loss + local_unsup_loss * self.learning_rate
            return (loss * weights).mean()
        else:
            return (sup_loss * weights).mean()

    def sup_loss(self, inputs, labels):
        if self.mode == 'regression':
            out = self.model(inputs)
            sup_loss = F.mse_loss(out, labels)
        elif self.mode == 'classification':
            out = self.model(inputs)
            out = F.softmax(out, dim=2)
            sup_loss = self.class_loss(out, labels)
        return sup_loss

    def local_unsup_loss(self, inputs):
        if self.task == 'semisupervised':
            y, M = self.components['unsup_encoder'](inputs)
        else:
            y, M = self.components['encoder'](inputs)
        g_enc = self.components['global_d'](y)
        l_enc = self.components['local_d'](M)

        if self.local:
            loss = self.localloss(l_enc, g_enc, inputs.graph_index)

        return loss

    def global_unsup_loss(self, inputs, labels, weights):
        y, M = self.components['encoder'](inputs)
        y_, M_ = self.components['unsup_encoder'](inputs)

        g_enc = self.components['ff1'](y)
        g_enc1 = self.components['ff2'](y_)

        loss = self.globalloss(g_enc, g_enc1)
        return loss

    def _prepare_batch(self, batch):
        """
        Prepares the batch for the model by converting the GraphData numpy arrays to torch tensors and moving them to the device.
        """
        inputs, labels, weights = batch
        inputs = BatchGraphData(inputs[0]).numpy_to_torch(self.device)

        _, labels, weights = super()._prepare_batch(([], labels, weights))

        if (len(labels) != 0) and (len(weights) != 0):
            labels = labels[0]
            weights = weights[0]

        return inputs, labels, weights

    def default_generator(
            self,
            dataset: dc.data.Dataset,
            epochs: int = 1,
            mode: str = 'fit',
            deterministic: bool = True,
            pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if self.mode == 'classification' and y_b is not None:
                    y_b = to_one_hot(y_b.flatten(), self.num_classes).reshape(
                        -1, self.num_tasks, self.num_classes)
                yield ([X_b], [y_b], [w_b])



"""
Implementation of the Ferminet class in pytorch
"""

from typing import List, Optional
# import torch.nn as nn
from rdkit import Chem
import numpy as np
from deepchem.utils.molecule_feature_utils import ALLEN_ELECTRONEGATIVTY
from deepchem.utils.geometry_utils import compute_pairwise_distances
from deepchem.models.torch_models import TorchModel
from deepchem.models.losses import L2Loss
import deepchem.models.optimizers as optimizers
import torch

from deepchem.utils.electron_sampler import ElectronSampler

# TODO look for the loss function(Hamiltonian)


def test_f(x: np.ndarray) -> np.ndarray:
    # dummy function which can be passed as the parameter f. f gives the log probability
    # TODO replace this function with forward pass of the model in future
    return 2 * np.log(np.random.uniform(low=0, high=1.0, size=np.shape(x)[0]))


class Ferminet(torch.nn.Module):
    """Approximates the log probability of the wave function of a molecule system using DNNs.
  """

    def __init__(self,
                 nucleon_pos: torch.Tensor,
                 nuclear_charge: torch.Tensor,
                 spin: tuple,
                 inter_atom: torch.Tensor,
                 n_one: List = [256, 256, 256, 256],
                 n_two: List = [32, 32, 32, 32],
                 determinant: int = 16) -> None:
        """
    Parameters:
    -----------
    nucleon_pos: torch.Tensor
        Torch tensor containing nucleus information of the molecule
    nuclear_charge: torch.Tensor
        Torch tensor containing the number of electron for each atom in the molecule
    spin: tuple
        Tuple in the format of (up_spin, down_spin)
    inter_atom: torch.Tensor
        Torch tensor containing the pairwise distances between the atoms in the molecule
    n_one: List
      List of hidden units for the one-electron stream in each layer
    n_two: List
      List of hidden units for the two-electron stream in each layer
    determinant: int
      Number of determinants for the final solution
    """
        super(Ferminet, self).__init__()
        if len(n_one) != len(n_two):
            raise ValueError(
                "The number of layers in one-electron and two-electron stream should be equal"
            )
        else:
            self.layers = len(n_one)
        self.nucleon_pos = nucleon_pos
        self.determinant = determinant
        self.spin = spin
        self.inter_atom = inter_atom
        self.n_one = n_one
        self.n_two = n_two
        self.determinant = determinant


class FerminetModel(TorchModel):
    """A deep-learning based Variational Monte Carlo method [1]_ for calculating the ab-initio
    solution of a many-electron system.

    This model aims to calculate the ground state energy of a multi-electron system
    using a baseline solution as the Hartree-Fock. An MCMC technique is used to sample
    electrons and DNNs are used to caluclate the square magnitude of the wavefunction,
    in which electron-electron repulsions also are included in the calculation(in the
    form of Jastrow factor envelopes). The model requires only the nucleus' coordinates
    as input.

    This method is based on the following paper:

    References
    ----------
    .. [1] Spencer, James S., et al. Better, Faster Fermionic Neural Networks. arXiv:2011.07125, arXiv, 13 Nov. 2020. arXiv.org, http://arxiv.org/abs/2011.07125.

    Note
    ----
    This class requires pySCF to be installed.
    """

    def __init__(
        self,
        nucleon_coordinates: List[List],
        spin: int,
        ion_charge: int,
        seed: Optional[int] = None,
        batch_no: int = 10,
        pretrain=True,
    ):
        """
    Parameters:
    -----------
    nucleon_coordinates: List[List]
      A list containing nucleon coordinates as the values with the keys as the element's symbol.
    spin: int
      The total spin of the molecule system.
    ion_charge: int
      The total charge of the molecule system.
    seed_no: int, optional (default None)
      Random seed to use for electron initialization.
    batch_no: int, optional (default 10)
      Number of batches of the electron's positions to be initialized.

    Attributes:
    -----------
    nucleon_pos: np.ndarray
        numpy array value of nucleon_coordinates
    electron_no: np.ndarray
        Torch tensor containing electrons for each atom in the nucleus
    molecule: ElectronSampler
        ElectronSampler object which performs MCMC and samples electrons
    """
        self.nucleon_coordinates = nucleon_coordinates
        self.seed = seed
        self.batch_no = batch_no
        self.spin = spin
        self.ion_charge = ion_charge
        self.batch_no = batch_no

        no_electrons = []
        nucleons = []
        electronegativity = []

        table = Chem.GetPeriodicTable()
        index = 0
        for i in self.nucleon_coordinates:
            atomic_num = table.GetAtomicNumber(i[0])
            electronegativity.append([index, ALLEN_ELECTRONEGATIVTY[i[0]]])
            no_electrons.append([atomic_num])
            nucleons.append(i[1])
            index += 1

        self.electron_no: np.ndarray = np.array(no_electrons)
        charge: np.ndarray = self.electron_no.reshape(
            np.shape(self.electron_no)[0])
        self.nucleon_pos: np.ndarray = np.array(nucleons)
        electro_neg = np.array(electronegativity)

        # Initialization for ionic molecules
        if np.sum(self.electron_no) < self.ion_charge:
            raise ValueError("Given charge is not initializable")

        # Initialization for ionic molecules
        if self.ion_charge != 0:
            if len(nucleons
                  ) == 1:  # for an atom, directly the charge is applied
                self.electron_no[0][0] -= self.ion_charge
            else:  # for a multiatomic molecule, the most electronegative atom gets a charge of -1 and vice versa. The remaining charges are assigned in terms of decreasing(for anionic charge) and increasing(for cationic charge) electronegativity.
                electro_neg = electro_neg[electro_neg[:, 1].argsort()]
                if self.ion_charge > 0:
                    for iter in range(self.ion_charge):
                        self.electron_no[int(electro_neg[iter][0])][0] -= 1
                else:
                    for iter in range(-self.ion_charge):
                        self.electron_no[int(electro_neg[-1 - iter][0])][0] += 1

        total_electrons = np.sum(self.electron_no)

        if self.spin >= 0:
            self.up_spin = (total_electrons + 2 * self.spin) // 2
            self.down_spin = total_electrons - self.up_spin
        else:
            self.down_spin = (total_electrons - 2 * self.spin) // 2
            self.up_spin = total_electrons - self.down_spin

        if self.up_spin - self.down_spin != self.spin:
            raise ValueError("Given spin is not feasible")

        nucl = torch.from_numpy(self.nucleon_pos)
        model = Ferminet(nucl,
                         spin=(self.up_spin, self.down_spin),
                         nuclear_charge=torch.tensor(charge),
                         inter_atom=torch.tensor(
                             compute_pairwise_distances(self.nucleon_pos,
                                                        self.nucleon_pos)))

        self.molecule: ElectronSampler = ElectronSampler(
            batch_no=self.batch_no,
            central_value=self.nucleon_pos,
            seed=self.seed,
            f=lambda x: test_f(x),  # Will be replaced in successive PR
            steps=1000,
            steps_per_update=20
        )  # sample the electrons using the electron sampler
        self.molecule.gauss_initialize_position(
            self.electron_no)  # initialize the position of the electrons
        adam = optimizers.AdamW()
        super(FerminetModel, self).__init__(
            model, optimizer=adam,
            loss=L2Loss())  # will update the loss in successive PR

    def prepare_hf_solution(self, x: np.ndarray) -> np.ndarray:
        """Prepares the HF solution for the molecule system which is to be used in pretraining

        Parameters
        ----------
        x: np.ndarray
        Numpy array of shape (number of electrons,3), which indicates the sampled electron's positions

        Returns
        -------
        hf_value: np.ndarray
        Numpy array of shape (number of electrons, number of electrons ) where ith row & jth value corresponds to the ith hartree fock orbital at the jth electron's coordinate
    """
        try:
            import pyscf
        except ModuleNotFoundError:
            raise ImportError("This module requires pySCF")

        molecule = ""
        for i in range(len(self.nucleon_pos)):
            molecule = molecule + self.nucleon_coordinates[i][0] + " " + str(
                self.nucleon_coordinates[i][1][0]) + " " + str(
                    self.nucleon_coordinates[i][1][1]) + " " + str(
                        self.nucleon_coordinates[i][1][2]) + ";"
        mol = pyscf.gto.Mole(atom=molecule, basis='sto-3g')
        mol.parse_arg = False
        mol.unit = 'Bohr'
        mol.spin = (self.up_spin - self.down_spin)
        mol.charge = self.ion_charge
        mol.build(parse_arg=False)
        mf = pyscf.scf.RHF(mol)
        mf.kernel()

        coefficients_all = mf.mo_coeff[:, :mol.nelectron]
        # Get the positions of all the electrons
        electron_positions = mol.atom_coords()[:mol.nelectron]
        # Evaluate all molecular orbitals at the positions of all the electrons
        orbital_values = np.dot(mol.eval_gto("GTOval", electron_positions),
                                coefficients_all)
        return orbital_values
